/*
 *  linux/fs/supermount/dentry.c
 *
 *  Original version:
 *      Copyright (C) 1995
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *  Rewritten for kernel 2.4.21 (C) 2003 Andrey Borzenkov
 *                                       (arvidjaar@mail.ru)
 *
 */

#define S_DBG_TRACE_CURRENT S_DBG_TRACE_DENTRY
#include "supermount.h"

static int
supermount_d_revalidate(struct dentry *dentry, int flags)
{
	struct super_block *sb = dentry->d_sb;
	struct dentry *subd;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dentry=%s", dentry->d_name.name);

	rc = 0;
	/*
	 * FIXME
	 * we need two version of prevent)umount as well
	 */
	mnt = subfs_prevent_umount(sb, dentry->d_inode);
	if (!mnt)
		goto out;
	
	subd = get_subfs_dentry(dentry);

	if (IS_ERR(subd))
		goto allow_umount;

	rc = 1;
	if (subd->d_op && subd->d_op->d_revalidate)
		rc = subd->d_op->d_revalidate(subd, flags);

	dput(subd);
allow_umount:
	subfs_allow_umount(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

/*
 * d_delete is used to maintain the same state of dcache for both
 * subfs and superfs. Without it there is a small window when
 * superdentry is already unhashed but subdentry not. While it will
 * be unhashed in supermount_clear_inode we may hit this situation
 * in between. It is interpreted as bug by supermount currently.
 *
 * this runs under dcache lock so we try to not block or crash here
 */
static int
supermount_d_delete(struct dentry *dentry)
{
	struct supermount_dentry_info *sdi;
	struct dentry *subd;

	/* FIXME */
	/* we probably can't use trace here :( */

	sdi = dentry->d_fsdata;
	if (!sdi)
		return 1;

	subd = sdi->dentry;
	if (!subd)
		return 1;

	/* we can't use d_drop here */
	list_del_init(&subd->d_hash);
	return 1;
}

/*
 * this should be d_release but we must be sure subfs dentry
 * is released before subfs inode, so we are the last user
 * in clear_inode. We (ab-)use d_iput for it
 */
static void
supermount_d_iput(struct dentry *dentry, struct inode *inode)
{
	struct super_block *sb = dentry->d_sb;
	struct supermount_dentry_info *sdi;
	struct dentry *subd = 0;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dentry=%s inode=0x%08lx", dentry->d_name.name, (unsigned long)inode);

	subfs_lock(sb);
	sdi = dentry->d_fsdata;
	if (sdi) {
		subd = sdi->dentry;
		list_del_init(&sdi->list);
		sdi->dentry = 0;
		SUPERMOUNT_BUG_LOCKED_ON(sb, (subd == 0) ^ is_dentry_obsolete(dentry));
	}
	subfs_unlock(sb);

	if (subd) {
		/*
		 * For root it is possible to have usage count > 1. This
		 * happens during umount of superfs in kill_super. First
		 * it destroys super dcache and then calls put_super
		 * that does mntput on submnt that finally kills root
		 * dentry and inode.
		 */
		SUPERMOUNT_BUG_ON(subd->d_sb->s_root != subd && atomic_read(&subd->d_count) != 1);
		/*
		 * normally it is already dropped it in d_delete
		 * except in shrink_dcache
		 */
		d_drop(subd);
		dput(subd);
	}

	iput(inode);

	SUPERMOUNT_TRACE_LEAVE(sb);
}
	

static void
supermount_d_release(struct dentry *dentry)
{
	struct super_block *sb = dentry->d_sb;
	struct supermount_dentry_info *sdi;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dentry=%s", dentry->d_name.name);

	sdi = dentry->d_fsdata;
	if (sdi)
		kfree(sdi);

	SUPERMOUNT_TRACE_LEAVE(sb);
}
	
struct dentry_operations supermount_dops = {
	.d_revalidate	= supermount_d_revalidate,
	.d_delete	= supermount_d_delete,
	.d_iput		= supermount_d_iput,
	.d_release	= supermount_d_release,
};
