/*
 *  linux/fs/supermount/file.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/dir.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/dir.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *  Rewriten for kernel 2.4. (C) 2001 MandrakeSoft Inc.
 *                                    Juan Quintela (quintela@mandrakesoft.com)
 *  Rewritten for kernel 2.4.21 (C) 2003 Andrey Borzenkov
 *                                       (arvidjaar@mail.ru)
 *
 */

#define S_DBG_TRACE_CURRENT S_DBG_TRACE_FILE
#include "supermount.h"

#define DEFAULT_POLLMASK (POLLIN | POLLOUT | POLLRDNORM | POLLWRNORM)

static loff_t
supermount_llseek(struct file *file, loff_t offset, int origin)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	loff_t rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	if (subfile->f_op && subfile->f_op->llseek)
		rc = subfile->f_op->llseek(subfile, offset, origin);
	else
		rc = default_llseek(subfile, offset, origin);
	file->f_pos = subfile->f_pos;
	file->f_reada = subfile->f_reada;
	file->f_version = subfile->f_version;

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static ssize_t
supermount_read(struct file *file, char *buf, size_t count, loff_t * ppos)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = -EINVAL;
	if (subfile->f_op && subfile->f_op->read)
		rc = subfile->f_op->read(subfile, buf, count, ppos);
	if (rc > 0)
		file->f_pos = subfile->f_pos = *ppos;

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static ssize_t
supermount_write(struct file *file, const char *buf,
		 size_t count, loff_t * ppos)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = 0;
	if (subfile->f_op && subfile->f_op->write)
		rc = subfile->f_op->write(subfile, buf, count, ppos);
	if (rc > 0) {
		struct inode *subinode = subfile->f_dentry->d_inode;

		file->f_pos = subfile->f_pos = *ppos;
		file->f_mode = subfile->f_mode;
		file->f_dentry->d_inode->i_size = subinode->i_size;
		file->f_dentry->d_inode->i_blocks = subinode->i_blocks;
		file->f_dentry->d_inode->i_mode = subinode->i_mode;
	}

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

int
supermount_readdir(struct file *file, void *buf, filldir_t fill_fn)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = -ENOTDIR;
	if (subfile->f_op && subfile->f_op->readdir)
		rc = subfile->f_op->readdir(subfile, buf, fill_fn);
	if (!rc)
		file->f_pos = subfile->f_pos;

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static unsigned int
supermount_poll(struct file *file, struct poll_table_struct *table)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = DEFAULT_POLLMASK;
	if (subfile->f_op && subfile->f_op->poll)
		rc = subfile->f_op->poll(subfile, table);

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_ioctl(struct inode *inode, struct file *file,
		 unsigned int cmd, unsigned long arg)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	struct inode *subinode;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = -ENOTTY;
	subinode = subfile->f_dentry->d_inode;
	if (subfile->f_op && subfile->f_op->ioctl)
		rc = subfile->f_op->ioctl(subinode, subfile, cmd, arg);

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

int
supermount_open(struct inode *inode, struct file *file)
{
	struct super_block *sb = inode->i_sb;
	struct dentry *subdent;
	struct file *subfile;
	struct vfsmount *submnt;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = supermount_init_file_info(file);
	if (rc)
		goto out;

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	submnt = subfs_get_mnt(sb);
	if (!submnt)
		goto out;

	subdent = get_subfs_dentry(file->f_dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_submnt;
	
	rc = subfs_get_read_file(sb, file);
	if (rc)
		goto put_subdent;

	if (file->f_mode & FMODE_WRITE) {
		rc = subfs_get_write_file(sb, file);
		if (rc)
			goto put_read_access;
	}

	/*
	 * the following is used to simplify error processing. dentry_open
	 * automatically does mntput and dput in error case, this may result
	 * in subfs being destroyed
	 * We just make sure we need to do mntput exactly once here;
	 * additionally it guards against accidental remounting of subfs
	 * until we has cleaned up
	 */
	submnt = mntget(submnt);
	subdent = dget(subdent);

	subfile = dentry_open(subdent, submnt, file->f_flags);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto put_write_access;
	/*
	 * no need to do extra mntput and dput, it is done automatically in
	 * dentry_open on error
	 */

	rc = supermount_attach_subfile(file, subfile);
	if (rc)
		goto put_subfile;

	subfile->f_mode = file->f_mode;
	/*
	 * this is needed for mmap to work. In current model vm_area
	 * is associated with superfile; we never explicitly call
	 * any vm method with subfile as pointer. But many drivers
	 * attach private structures to this field and mmap of special
	 * files on supermount fs won't work without it
	 */
	file->private_data = subfile->private_data;

	rc = 0;

	/*
	 * Now get rid of extra mntget and dget
	 */
	goto put_subdent;

	/*
	 * error cleanup
	 */

put_subfile:
	fput(subfile);
put_write_access:
	subfs_put_write_file(sb, file);
put_read_access:
	subfs_put_read_file(sb, file);
put_subdent:
	dput(subdent);
put_submnt:
	mntput(submnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_flush(struct file *file)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = 0;
	if (subfile->f_op && subfile->f_op->flush)
		rc = subfile->f_op->flush(subfile);

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

/*
 * if subfile is NULL it has already been released in supermount_clean_files
 * together with adjusting open/write counters. Else we do it here.
 *
 * The reason is, it may be called long after media has been changed
 * and we definitely do not want this function to mess up the
 * new subfs state.
 */
static int
supermount_release(struct inode *inode, struct file *file)
{
	struct file *subfile = 0;
	struct super_block *sb = inode->i_sb;
	struct supermount_file_info *sfi;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	/*
	 * this sucks. But there does not seem to be any way
	 * to distinguish between ENOMEM on _open (legitimate
	 * case) and anything else (plain bug)
	 */
	sfi = file->f_supermount;
	if (sfi) {
		list_del(&sfi->list);
		subfile = sfi->file;
		sfi->file = 0;
		SUPERMOUNT_BUG_LOCKED_ON(sb, (subfile == 0) ^ is_file_obsolete(file));
	} else
		supermount_warning(sb, "no supermount file info attached");
	subfs_unlock(sb);

	if (subfile) {
		int bug = atomic_read(&subfile->f_count) != 1;
		fput(subfile);
		if (file->f_mode & FMODE_WRITE)
			subfs_put_write_file(sb, file);
		subfs_put_read_file(sb, file);
		SUPERMOUNT_BUG_ON(bug);
	}

	kfree(sfi);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return 0;

}

static int
supermount_fsync(struct file *file, struct dentry *dentry, int datasync)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = -EINVAL;
	if (subfile->f_op && subfile->f_op->fsync)
		rc = subfile->f_op->fsync(subfile, subfile->f_dentry, datasync);

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_fasync(int fd, struct file *file, int on)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = -EINVAL;
	if (subfile->f_op && subfile->f_op->fasync)
		rc = subfile->f_op->fasync(fd, subfile, on);

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_lock(struct file *file, int cmd, struct file_lock *fl)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct file *subfile;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = -ESTALE;
	if (subfs_check_disk_change(sb))
		goto out;

	subfile = get_subfs_file(file);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto out;

	rc = 0;
	if (subfile->f_op && subfile->f_op->lock)
		rc = subfile->f_op->lock(subfile, cmd, fl);
	else if (cmd == F_GETLK)
		posix_test_lock(file, fl);

	fput(subfile);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

/* Fixme:
 * readv: easy, export churnk from vfs
 * writev: easy, export churnk from vfs
 * sendpage: only used for networking, not needed
 * get_unmmapped_area: only used for devices, not needed
 */

struct file_operations supermount_dir_operations = {
	.llseek		= supermount_llseek,
	.read		= supermount_read,
	.readdir	= supermount_readdir,
	.ioctl		= supermount_ioctl,
	.open		= supermount_open,
	.flush		= supermount_flush,
	.release	= supermount_release,
	.fsync		= supermount_fsync,
	.fasync		= supermount_fasync,
};

struct file_operations supermount_file_operations = {
	.llseek		= supermount_llseek,
	.read		= supermount_read,
	.write		= supermount_write,
	.poll		= supermount_poll,
	.ioctl		= supermount_ioctl,
	.mmap		= supermount_file_mmap, /* from filemap.c */
	.open		= supermount_open,
	.flush		= supermount_flush,
	.release	= supermount_release,
	.fsync		= supermount_fsync,
	.fasync		= supermount_fasync,
	.lock		= supermount_lock,
};
