/*
 *  linux/fs/supermount/namei.c
 *
 *  Original version:
 *      Copyright (C) 1995
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/namei.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/namei.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *  Rewritten for kernel 2.4.21 (C) 2003 Andrey Borzenkov
 *                                       (arvidjaar@mail.ru)
 *
 */

#define S_DBG_TRACE_CURRENT S_DBG_TRACE_NAMEI
#include "supermount.h"

struct inode *
supermount_create_inode(struct super_block *sb)
{
	struct inode *inode = new_inode(sb);
	struct supermount_inode_info *sii;
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	SUPERMOUNT_TRACE_ENTER(sb);

	if (!inode)
		goto out;

	sii = kmalloc(sizeof(*sii), GFP_KERNEL);
	if (!sii) {
		inode = 0;
		goto out;
	}

	inode->i_mode = sbi->s_default_mode;
	if (sb->s_root) {
		inode->i_uid = sb->s_root->d_inode->i_uid;
		inode->i_gid = sb->s_root->d_inode->i_gid;
	} else {
		inode->i_uid = 0;
		inode->i_gid = 0;
	}
	inode->i_blksize = inode->i_sb->s_blocksize;
	inode->i_rdev = NODEV;
	inode->i_atime = inode->i_mtime = inode->i_ctime = CURRENT_TIME;

	memset(sii, 0, sizeof(*sii));
	INIT_LIST_HEAD(&sii->list);
	sii->host = inode;
	sii->generation = supermount_sb_generation(sb);
	inode->u.generic_ip = sii;

out:
	SUPERMOUNT_TRACE_LEAVE_ARGS(sb, "inode=0x%08lx", (unsigned long)inode);

	return inode;
}

static void
supermount_update_super_inode(struct super_block *sb, struct inode *superi, struct inode *subi)
{
	struct supermount_inode_info *sii = supermount_i(superi);

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "superi=0x%08lx subi=0x%08lx", (unsigned long)superi, (unsigned long)subi);

	sii->inode = igrab(subi);

	superi->i_mode = subi->i_mode;
	superi->i_uid = subi->i_uid;
	superi->i_gid = subi->i_gid;
	superi->i_nlink = subi->i_nlink;
	superi->i_size = subi->i_size;
	superi->i_atime = subi->i_atime;
	superi->i_ctime = subi->i_ctime;
	superi->i_mtime = subi->i_mtime;
	superi->i_blksize = subi->i_blksize;
	superi->i_blocks = subi->i_blocks;
	superi->i_rdev = subi->i_rdev;
	superi->i_version++;

	if (S_ISDIR(superi->i_mode)) {
		superi->i_op = &supermount_dir_iops;
		superi->i_fop = &supermount_dir_operations;
	} else if (S_ISLNK(superi->i_mode)) {
		superi->i_op = &supermount_symlink_iops;
		superi->i_mapping = subi->i_mapping;
	} else {
		superi->i_op = &supermount_file_iops;
		superi->i_fop = &supermount_file_operations;
		superi->i_mapping = subi->i_mapping;
	}

	SUPERMOUNT_TRACE_LEAVE(sb);

}

/*
 * Attach superinode to subinode and instantiate dentry with this inode
 * this needs lock as otherweise we can end up adding stale inode
 */

static int
attach_and_prepare(struct super_block *sb, struct dentry *dentry, struct dentry *subd)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct supermount_inode_info *sii;
	struct inode *superi = 0;
	struct inode *subi = subd->d_inode;
	int rc;
	struct list_head *ptr;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);

	SUPERMOUNT_BUG_LOCKED_ON(sb, !subi);

	rc = supermount_init_dentry_info(dentry);
	if (rc)
		goto drop_subd;

	list_for_each(ptr, &sbi->s_inodes) {
		sii = list_entry(ptr, struct supermount_inode_info, list);

		if (sii->inode == subi) {
			superi = igrab(sii->host);
			break;
		}
	}

	if (!superi) {
		superi = supermount_create_inode(sb);
		if (!superi)
			goto drop_subd;
		supermount_update_super_inode(sb, superi, subi);
		supermount_list_add_inode(superi);
	}

	supermount_attach_subdent(dentry, subd);
	d_add(dentry, superi);

	/*
	 * FIXME
	 * at this point subi and subd must be exact copy with the only
	 * difference that refcount is +1 for subfs. Just for now we skip
	 * a BUG due to the lack of time to test
	 *
	 * note that it may need dcache lock and possibly inode lock
	 */

	if (atomic_read(&dentry->d_count) + 1 != atomic_read(&subd->d_count))
		supermount_error(sb, "dentry refcount mismatch");

	if (atomic_read(&superi->i_count) + 1 != atomic_read(&subi->i_count))
		supermount_error(sb, "inode refcount mismatch");

	rc = 0;
	goto out;

drop_subd:
	d_drop(subd);
out:
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
	/*
	 * subdent is implicitly freed on return if we skip dget here
	 */
}

static inline void
update_nlink(struct inode *superi, struct inode *subi)
{
	superi->i_nlink = subi->i_nlink;
}

static inline void
update_ctime(struct inode *superi, struct inode *subi)
{
	superi->i_ctime = subi->i_ctime;
}

static inline void
update_mtime(struct inode *superi, struct inode *subi)
{
	superi->i_mtime = subi->i_mtime;
}

/*
 * FIXME
 * this is copied over from inode.c:update_atime. Currently
 * we do not sync inode times between subfs and superfs, so these
 * checks are needed to ensure subfs atime is not set to "stale"
 * value
 *
 * This will go away as soon as somebody goes through code and makes
 * sure both inodes always have the same time
 */

static void
supermount_update_atime(struct inode *inode, struct inode *subi)
{
        if ( IS_NOATIME (inode) ) return;
        if ( IS_NODIRATIME (inode) && S_ISDIR (inode->i_mode) ) return;
        if ( IS_RDONLY (inode) ) return;
        subi->i_atime = inode->i_atime;
        mark_inode_dirty_sync (subi);
}


/* inode methods */

static int
supermount_create(struct inode *dir, struct dentry *dentry, int mode)
{
	struct super_block *sb = dir->i_sb;
	struct dentry *subdent;
	struct inode *subdir;
	struct dentry *parent;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dir=0x%08lx dentry=%s", (unsigned long)dir, dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dir);
	if (rc)
		goto go_offline;

	subdir = get_subfs_inode(dir);
	rc = PTR_ERR(subdir);
	if (IS_ERR(subdir))
		goto put_read_access;

	parent = get_subfs_dentry(dentry->d_parent);
	rc = PTR_ERR(parent);
	if (IS_ERR(parent))
		goto put_subdir;

	rc = -EACCES;
	if (!subdir->i_op || !subdir->i_op->create)
		goto put_parent;
	
	subdent = new_subfs_dentry(parent, dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_parent;

	rc = subfs_get_write_inode(sb, dir);
	if (rc)
		goto put_subdent;

	rc = subdir->i_op->create(subdir, subdent, mode);
	if (rc)
		goto put_write_access;

	rc = attach_and_prepare(sb, dentry, subdent);

put_write_access:
	subfs_put_write_inode(sb, dir);
put_subdent:
	dput(subdent);
put_parent:
	dput(parent);
put_subdir:
	iput(subdir);
put_read_access:
	subfs_put_read_inode(sb, dir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

/*
 * When we come here, dentry has just been allocated and not yet hashed.
 * If subfs lookup fails, we just return the same dentry with NULL inode
 * that is interpreted as "no such file" and next dput will destroy dentry.
 * For subfs we make sure it is not hashed as well. This ensures that no
 * negative dentries ever exist (because it impossible to revalidate them)
 */
static struct dentry *
supermount_lookup(struct inode *dir, struct dentry *dentry)
{
	struct super_block *sb = dir->i_sb;
	struct dentry *rc, *subdent, *parent;
	struct inode *subdir;
	struct vfsmount *mnt;
	int ret;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dir=0x%08lx dentry=%s", (unsigned long)dir, dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = (struct dentry *)mnt;
	if (IS_ERR(mnt))
		goto out;

	ret = subfs_get_read_inode(sb, dir);
	rc = ERR_PTR(ret);
	if (ret)
		goto go_offline;

	subdir = get_subfs_inode(dir);
	rc = (struct dentry *)subdir;
	if (IS_ERR(subdir))
		goto put_read_access;

	parent = get_subfs_dentry(dentry->d_parent);
	rc = parent;
	if (IS_ERR(parent))
		goto put_subdir;

	subdent = new_subfs_dentry(parent, dentry);
	rc = subdent;
	if (IS_ERR(subdent))
		goto put_parent;

	rc = subdir->i_op->lookup(subdir, subdent);
	if (rc)
		goto put_subdent;

	if (subdent->d_inode)
		rc = ERR_PTR(attach_and_prepare(sb, dentry, subdent));
	else
		/*
		 * make sure negative dentry is not cached
		 */
		d_drop(subdent);

put_subdent:
	dput(subdent);
put_parent:
	dput(parent);
put_subdir:
	iput(subdir);
put_read_access:
	subfs_put_read_inode(sb, dir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_link(struct dentry *old_dentry, struct inode *dir,
		struct dentry *new_dentry)
{
	struct super_block *sb = dir->i_sb;
	struct dentry *old_subdent, *new_parent, *new_subdent;
	struct inode *subdir;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "from=%s dir=0x%08lx to=%s", old_dentry->d_name.name, (unsigned long)dir, new_dentry->d_name.name);

	SUPERMOUNT_BUG_ON(new_dentry->d_inode);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dir);
	if (rc)
		goto go_offline;

	subdir = get_subfs_inode(dir);
	rc = PTR_ERR(subdir);
	if (IS_ERR(subdir))
		goto put_read_access;

	new_parent = get_subfs_dentry(new_dentry->d_parent);
	rc = PTR_ERR(new_parent);
	if (IS_ERR(new_parent))
		goto put_subdir;

	rc = -EPERM;
	if (!subdir->i_op || !subdir->i_op->link)
		goto put_new_parent;

	old_subdent = get_subfs_dentry(old_dentry);
	rc = PTR_ERR(old_subdent);
	if (IS_ERR(old_subdent))
		goto put_new_parent;

	new_subdent = new_subfs_dentry(new_parent, new_dentry);
	rc = PTR_ERR(new_subdent);
	if (IS_ERR(new_subdent))
		goto put_old_subdent;

	rc = subfs_get_write_inode(sb, dir);
	if (rc)
		goto put_new_subdent;

	rc = subdir->i_op->link(old_subdent, subdir, new_subdent);
	if (rc)
		goto put_write_access;

	rc = attach_and_prepare(sb, new_dentry, new_subdent);

put_write_access:
	subfs_put_write_inode(sb, dir);
put_new_subdent:
	dput(new_subdent);
put_old_subdent:
	dput(old_subdent);
put_new_parent:
	dput(new_parent);
put_subdir:
	iput(subdir);
put_read_access:
	subfs_put_read_inode(sb, dir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_unlink(struct inode *dir, struct dentry *dentry)
{
	struct super_block *sb = dir->i_sb;
	struct dentry *parent, *subdent;
	struct inode *subdir;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dir=0x%08lx dentry=%s", (unsigned long)dir, dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dir);
	if (rc)
		goto go_offline;

	subdir = get_subfs_inode(dir);
	rc = PTR_ERR(subdir);
	if (IS_ERR(subdir))
		goto put_read_access;

	parent = get_subfs_dentry(dentry->d_parent);
	rc = PTR_ERR(parent);
	if (IS_ERR(parent))
		goto put_subdir;

	rc = -EPERM;
	if (!subdir->i_op || !subdir->i_op->unlink)
		goto put_parent;

	subdent = get_subfs_dentry(dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_parent;

	/*
	 * below is not a typo. We have to mark _deleted_ inode
	 * for possible later delete in clear_inode
	 */
	rc = subfs_get_write_inode(sb, dentry->d_inode);
	if (rc)
		goto put_subdent;


	rc = subdir->i_op->unlink(subdir, subdent);
	if (rc)
		goto put_write_access;
	/*
	 * we never do d_delete; supermount cannot handle negative
	 * dentries
	 */
	d_drop(subdent);
	d_drop(dentry);

	update_nlink(dir, subdir);
	update_nlink(dentry->d_inode, subdent->d_inode);
	update_ctime(dir, subdir);
	update_mtime(dir, subdir);
	update_ctime(dentry->d_inode, subdent->d_inode);

put_write_access:
	/*
	 * we can't put write access if there are pending deletes
	 * so we leave it on and let it be put in clear_inode for
	 * i_nlink == 0
	 */
	if (dentry->d_inode->i_nlink)
		subfs_put_write_inode(sb, dentry->d_inode);
put_subdent:
	dput(subdent);
put_parent:
	dput(parent);
put_subdir:
	iput(subdir);
put_read_access:
	subfs_put_read_inode(sb, dir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_symlink(struct inode *dir, struct dentry *dentry,
		   const char *symname)
{
	struct super_block *sb = dir->i_sb;
	struct dentry *subdent, *parent;
	struct inode *subdir;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dir=0x%08lx dentry=%s", (unsigned long)dir, dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dir);
	if (rc)
		goto go_offline;

	subdir = get_subfs_inode(dir);
	rc = PTR_ERR(subdir);
	if (IS_ERR(subdir))
		goto put_read_access;

	parent = get_subfs_dentry(dentry->d_parent);
	rc = PTR_ERR(parent);
	if (IS_ERR(parent))
		goto put_subdir;

	rc = -EPERM;
	if (!subdir->i_op || !subdir->i_op->symlink)
		goto put_parent;

	subdent = new_subfs_dentry(parent, dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_parent;

	rc = subfs_get_write_inode(sb, dir);
	if (rc)
		goto put_subdent;

	rc = subdir->i_op->symlink(subdir, subdent, symname);
	if (rc)
		goto put_write_access;

	rc = attach_and_prepare(sb, dentry, subdent);

put_write_access:
	subfs_put_write_inode(sb, dir);
put_subdent:
	dput(subdent);
put_parent:
	dput(parent);
put_subdir:
	iput(subdir);
put_read_access:
	subfs_put_read_inode(sb, dir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct super_block *sb = dir->i_sb;
	struct dentry *subdent, *parent;
	struct inode *subdir;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dir=0x%08lx dentry=%s", (unsigned long)dir, dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dir);
	if (rc)
		goto go_offline;

	subdir = get_subfs_inode(dir);
	rc = PTR_ERR(subdir);
	if (IS_ERR(subdir))
		goto put_read_access;

	parent = get_subfs_dentry(dentry->d_parent);
	rc = PTR_ERR(parent);
	if (IS_ERR(parent))
		goto put_subdir;

	rc = -EPERM;
	if (!subdir->i_op || !subdir->i_op->mkdir)
		goto put_parent;

	subdent = new_subfs_dentry(parent, dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_parent;

	rc = subfs_get_write_inode(sb, dir);
	if (rc)
		goto put_subdent;

	rc = subdir->i_op->mkdir(subdir, subdent, mode);
	if (rc)
		goto put_write_access;

	update_nlink(dir, subdir);
	rc = attach_and_prepare(sb, dentry, subdent);

put_write_access:
	subfs_put_write_inode(sb, dir);
put_subdent:
	dput(subdent);
put_parent:
	dput(parent);
put_subdir:
	iput(subdir);
put_read_access:
	subfs_put_read_inode(sb, dir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct super_block *sb = dir->i_sb;
	struct dentry *subdent, *parent;
	struct inode *subdir;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dir=0x%08lx dentry=%s", (unsigned long)dir, dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dir);
	if (rc)
		goto go_offline;

	subdir = get_subfs_inode(dir);
	rc = PTR_ERR(subdir);
	if (IS_ERR(subdir))
		goto put_read_access;

	rc = -EPERM;
	if (!subdir->i_op || !subdir->i_op->rmdir)
		goto put_subdir;

	parent = get_subfs_dentry(dentry->d_parent);
	rc = PTR_ERR(parent);
	if (IS_ERR(parent))
		goto put_subdir;

	subdent = get_subfs_dentry(dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_parent;

	/* cf. supermount_unlink */
	rc = subfs_get_write_inode(sb, dentry->d_inode);
	if (rc)
		goto put_subdent;

	rc = subdir->i_op->rmdir(subdir, subdent);
	if (rc)
		goto put_write_access;

	/*
	 * cf. supermount_unlink
	 */
	d_drop(subdent);
	/*
	 * there is no use to d_drop dentry because it is already
	 * unhashed; but let's not rely on it
	 */
	d_drop(dentry);

	update_nlink(dir, subdir);
	update_nlink(dentry->d_inode, subdent->d_inode);
	update_ctime(dir, subdir);
	update_mtime(dir, subdir);
	update_ctime(dentry->d_inode, subdent->d_inode);

put_write_access:
	/* cf. supermount_unlink */
	if (dentry->d_inode->i_nlink)
		subfs_put_write_inode(sb, dentry->d_inode);
put_subdent:
	dput(subdent);
put_parent:
	dput(parent);
put_subdir:
	iput(subdir);
put_read_access:
	subfs_put_read_inode(sb, dir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_mknod(struct inode *dir, struct dentry *dentry, int
		 mode, int dev)
{
	struct super_block *sb = dir->i_sb;
	struct dentry *subdent, *parent;
	struct inode *subdir;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dir=0x%08lx dentry=%s", (unsigned long)dir, dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dir);
	if (rc)
		goto go_offline;

	subdir = get_subfs_inode(dir);
	rc = PTR_ERR(subdir);
	if (IS_ERR(subdir))
		goto put_read_access;

	rc = -EPERM;
	if (!subdir->i_op || !subdir->i_op->mknod)
		goto put_subdir;

	parent = get_subfs_dentry(dentry->d_parent);
	rc = PTR_ERR(parent);
	if (IS_ERR(parent))
		goto put_subdir;

	subdent = new_subfs_dentry(parent, dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_parent;

	rc = subfs_get_write_inode(sb, dir);
	if (rc)
		goto put_subdent;

	rc = subdir->i_op->mknod(subdir, subdent, mode, dev);
	if (rc)
		goto put_write_access;

	rc = attach_and_prepare(sb, dentry, subdent);

put_write_access:
	subfs_put_write_inode(sb, dir);
put_subdent:
	dput(subdent);
put_parent:
	dput(parent);
put_subdir:
	iput(subdir);
put_read_access:
	subfs_put_read_inode(sb, dir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_rename(struct inode *olddir, struct dentry *olddentry,
		  struct inode *newdir, struct dentry *newdentry)
{
	struct super_block *sb = olddir->i_sb;
	struct dentry *oldsubdent, *newsubdent, *newparent;
	struct inode *oldsubdir, *newsubdir;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "olddir=0x%08lx olddentry=%s newdir=0x%08lx newdentry=%s", (unsigned long)olddir, olddentry->d_name.name, (unsigned long)newdir, newdentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, olddir);
	if (rc)
		goto go_offline;

	oldsubdir = get_subfs_inode(olddir);
	rc = PTR_ERR(oldsubdir);
	if (IS_ERR(oldsubdir))
		goto put_read_access;

	rc = -EPERM;
	if (!oldsubdir->i_op || !oldsubdir->i_op->rename)
		goto put_old_subdir;

	oldsubdent = get_subfs_dentry(olddentry);
	rc = PTR_ERR(oldsubdent);
	if (IS_ERR(oldsubdent))
		goto put_old_subdir;

	newsubdir = get_subfs_inode(newdir);
	rc = PTR_ERR(newsubdir);
	if (IS_ERR(newsubdir))
		goto put_old_subdent;

	newparent = get_subfs_dentry(newdentry->d_parent);
	rc = PTR_ERR(newparent);
	if (IS_ERR(newparent))
		goto put_new_subdir;

	/*
	 * Heck, this can be both active and negative dentry
	 */
	if (newdentry->d_inode)
		newsubdent = get_subfs_dentry(newdentry);
	else
		newsubdent = new_subfs_dentry(newparent, newdentry);
	rc = PTR_ERR(newsubdent);
	if (IS_ERR(newsubdent))
		goto put_new_parent;

	rc = subfs_get_write_inode(sb, olddir);
	if (rc)
		goto put_new_subdent;

	rc = oldsubdir->i_op->rename(oldsubdir, oldsubdent,
				     newsubdir, newsubdent);
	if (rc)
		goto put_write_access;

	/*
	 * FIXME
	 * this is workaround for negative dentry problem in d_move. Because
	 * we do not hash them, attempt to move old dentry into hash
	 * queue of new dentry fails. So we hash it (it will go away in
	 * several moments)
	 *
	 * Unfortunately hashing them is not trivial because we (ab-)use
	 * d_iput to handle subdentry; no inode - no d_iput. And so far
	 * dcache does not allow any other place to catch dentry removal.
	 * It may be possible to use d_release for that in case of negative
	 * dentry.
	 *
	 * It is likely impossible to _not_ hash subfs dentries because
	 * it makes clean up on umount quite questionable
	 */
	if (list_empty(&newdentry->d_hash))
		d_rehash(newdentry);
	if (list_empty(&newsubdent->d_hash))
		d_rehash(newsubdent);

	/*
	 * this is strictly speaking conditional on FS_ODD_RENAME
	 * flag, but as of this writing this flag is set only
	 * for NFS or intermezzo and it hopefully goes away sometimes ...
	 *
	 * Supermount patch adds code to do_kern_mount that checks
	 * for this flag as refuses mounting
	 */
	d_move(oldsubdent, newsubdent);

	update_nlink(olddir, oldsubdir);
	update_nlink(newdir, newsubdir);
	update_nlink(olddentry->d_inode, oldsubdent->d_inode);
	if (newdentry->d_inode)
		update_nlink(newdentry->d_inode, newsubdent->d_inode);
	update_ctime(olddir, oldsubdir);
	update_mtime(olddir, oldsubdir);
	update_ctime(olddentry->d_inode, oldsubdent->d_inode);

put_write_access:
	/* cf. supermount_unlink */
	if (newdentry->d_inode && newdentry->d_inode->i_nlink)
		subfs_put_write_inode(sb, olddir);
put_new_subdent:
	dput(newsubdent);
put_new_parent:
	dput(newparent);
put_new_subdir:
	iput(newsubdir);
put_old_subdent:
	dput(oldsubdent);
put_old_subdir:
	iput(oldsubdir);
put_read_access:
	subfs_put_read_inode(sb, olddir);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int 
supermount_readlink(struct dentry *dentry, char *buffer , int buflen)
{
	struct super_block *sb = dentry->d_sb;
	struct dentry *subdent;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dentry=%s", dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dentry->d_inode);
	if (rc)
		goto go_offline;

	subdent = get_subfs_dentry(dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_read_access;

	rc = -EINVAL;
	if (!subdent->d_inode->i_op || !subdent->d_inode->i_op->readlink)
		goto put_subdent;

	SUPERMOUNT_UPDATE_ATIME(sb,dentry->d_inode,subdent->d_inode);
	rc = subdent->d_inode->i_op->readlink(subdent, buffer, buflen);

put_subdent:
	dput(subdent);
put_read_access:
	subfs_put_read_inode(sb, dentry->d_inode);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	struct super_block *sb = dentry->d_sb;
	struct dentry *subdent;
	int rc;
	struct vfsmount *mnt;
	
	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dentry=%s", dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dentry->d_inode);
	if (rc)
		goto go_offline;

	subdent = get_subfs_dentry(dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_read_access;

	SUPERMOUNT_UPDATE_ATIME(sb,dentry->d_inode,subdent->d_inode);
	rc = subdent->d_inode->i_op->follow_link(subdent, nd);

	dput(subdent);	
put_read_access:
	subfs_put_read_inode(sb, dentry->d_inode);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_permission(struct inode *inode, int mask)
{
	struct super_block *sb = inode->i_sb;
	struct inode *subi;
	int rc;
	int need_write_access;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "inode=0x%08lx", (unsigned long)inode);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, inode);
	if (rc)
		goto go_offline;

	subi = get_subfs_inode(inode);
	rc = PTR_ERR(subi);
	if (IS_ERR(subi))
		goto put_read_access;

	need_write_access = !(sb->s_flags & MS_RDONLY) && (mask & MAY_WRITE) && IS_RDONLY(subi);
	if (need_write_access) {
		rc = subfs_get_write_inode(sb, inode);
		if (rc)
			goto put_subi;
	}

	if (subi->i_op && subi->i_op->permission)
		rc = subi->i_op->permission(subi, mask);
	else
		rc = vfs_permission(subi, mask);

	if (need_write_access)
		subfs_put_write_inode(sb, inode);

put_subi:
	iput(subi);
put_read_access:
	subfs_put_read_inode(sb, inode);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_setattr(struct dentry *dentry, struct iattr *attr)
{
	struct super_block *sb = dentry->d_sb;
	struct inode *subi;
	struct dentry *subdent;
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dentry=%s", dentry->d_name.name);

	mnt = subfs_go_online(sb);
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	rc = subfs_get_read_inode(sb, dentry->d_inode);
	if (rc)
		goto go_offline;

	subdent = get_subfs_dentry(dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_read_access;

	rc = subfs_get_write_inode(sb, dentry->d_inode);
	if (rc)
		goto put_subdent;

	subi = subdent->d_inode;
	if (subi->i_op && subi->i_op->setattr)
		rc = subi->i_op->setattr(subdent, attr);
	else {
		rc = inode_change_ok(subi, attr);
		/*
		 * FIXME
		 * What to do with quota?
		 */
		if (!rc)
			rc = inode_setattr(subi, attr);
	}

	if (rc)
		goto put_write_access;

	/* 
	 * If it worked, then we need to mark the modification
	 * to the subfs, and we also need to propogate the
	 * change up to the shadowing inode.  
	 */
	attr->ia_mode = subi->i_mode;
	attr->ia_uid = subi->i_uid;
	attr->ia_gid = subi->i_gid;
	attr->ia_size = subi->i_size;
	attr->ia_atime = subi->i_atime;
	attr->ia_mtime = subi->i_mtime;
	attr->ia_ctime = subi->i_ctime;
	attr->ia_valid =
	    ATTR_UID | ATTR_GID | ATTR_MODE | ATTR_SIZE |
	    ATTR_ATIME | ATTR_MTIME | ATTR_CTIME;
	inode_setattr(dentry->d_inode, attr);

put_write_access:
	subfs_put_write_inode(sb, dentry->d_inode);
put_subdent:
	dput(subdent);
put_read_access:
	subfs_put_read_inode(sb, dentry->d_inode);
go_offline:
	subfs_go_offline(sb, mnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_getattr(struct vfsmount *mnt, struct dentry *dentry,
		   struct kstat *stat)
{
	struct super_block *sb = dentry->d_sb;
	struct dentry *subdent;
	int rc;
	struct vfsmount *submnt;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dentry=%s", dentry->d_name.name);

	submnt = subfs_go_online(sb);
	rc = PTR_ERR(submnt);
	if (IS_ERR(submnt))
		goto out;

	rc = subfs_get_read_inode(sb, dentry->d_inode);
	if (rc)
		goto go_offline;

	subdent = get_subfs_dentry(dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto put_read_access;

	if (subdent->d_inode->i_op && subdent->d_inode->i_op->getattr)
		rc = subdent->d_inode->i_op->getattr(submnt, subdent, stat);
	else
		rc = getattr_full(submnt, subdent, stat);

	dput(subdent);
put_read_access:
	subfs_put_read_inode(sb, dentry->d_inode);
go_offline:
	subfs_go_offline(sb, submnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_root_getattr(struct vfsmount *mnt, struct dentry *dentry,
			struct kstat *stat)
{
	struct super_block *sb = dentry->d_sb;
	int retval;

	SUPERMOUNT_TRACE_ENTER_ARGS(sb, "dentry=%s", dentry->d_name.name);

	/* the lock is to ensure subs_is_active is atomic */
	subfs_lock(sb);
	if (subfs_is_active(sb)) {
		subfs_unlock(sb);
		retval = supermount_getattr(mnt, dentry, stat);
	} else {
		retval = getattr_full(mnt, dentry, stat);
		subfs_unlock(sb);
	}


	SUPERMOUNT_TRACE_LEAVE(sb);

	return retval;
}

/*
 * directories can handle most operations...  supermount/namei.c just
 * passes them through to the underlying subfs, except for lookup().
 */

/* truncate: is not necesary, handled with setattr
 * revalidate: only needed by nfs
 * FixMe: implement accl functions
 */
struct inode_operations supermount_root_iops = {
	.create		= supermount_create,
	.lookup		= supermount_lookup,
	.link		= supermount_link,
	.unlink		= supermount_unlink,
	.symlink	= supermount_symlink,
	.mkdir		= supermount_mkdir,
	.rmdir		= supermount_rmdir,
	.mknod		= supermount_mknod,
	.rename		= supermount_rename,
	.permission	= supermount_permission,
	.setattr	= supermount_setattr,
	.getattr	= supermount_root_getattr,
};

struct inode_operations supermount_dir_iops = {
	.create		= supermount_create,
	.lookup		= supermount_lookup,
	.link		= supermount_link,
	.unlink		= supermount_unlink,
	.symlink	= supermount_symlink,
	.mkdir		= supermount_mkdir,
	.rmdir		= supermount_rmdir,
	.mknod		= supermount_mknod,
	.rename		= supermount_rename,
	.permission	= supermount_permission,
	.setattr	= supermount_setattr,
	.getattr	= supermount_getattr,
};

struct inode_operations supermount_symlink_iops = {
	.readlink	= supermount_readlink,
	.follow_link	= supermount_follow_link,
	.setattr	= supermount_setattr,
	.getattr	= supermount_getattr,
};

struct inode_operations supermount_file_iops = {
	.setattr	= supermount_setattr,
	.getattr	= supermount_getattr,
};
