/*
 *  linux/fs/supermount/super.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/super.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2, 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *  Rewritten for kernel 2.4.21 (C) 2003 Andrey Borzenkov
 *                                       (arvidjaar@mail.ru)
 *
 */

#define S_DBG_TRACE_CURRENT S_DBG_TRACE_SUBFS
#include "supermount.h"

static void __subfs_put_read_access(struct super_block *sb);

/*
 * close all open files on subfs
 */
static void
supermount_clean_files(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct list_head *ptr, *n;

	SUPERMOUNT_TRACE_ENTER(sb);

	list_for_each_safe(ptr, n, &sbi->s_files) {
		struct supermount_file_info *sfi;
		struct file *subfile;

		sfi = list_entry(ptr, struct supermount_file_info, list);

		subfile = sfi->file;
		sfi->file = 0;
		list_del_init(&sfi->list);

		SUPERMOUNT_BUG_LOCKED_ON(sb, !subfile);

		fput(subfile);

		if (sfi->host->f_mode & FMODE_WRITE)
			sbi->s_writecount--;
		__subfs_put_read_access(sb);
	}

	SUPERMOUNT_TRACE_LEAVE(sb);

}

/*
 * close all open dentries on subfs
 */
static void
supermount_clean_dentries(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct list_head *ptr, *n;

	SUPERMOUNT_TRACE_ENTER(sb);

	list_for_each_safe(ptr, n, &sbi->s_dentries) {
		struct supermount_dentry_info *sdi;
		struct dentry *subdent;

		sdi = list_entry(ptr, struct supermount_dentry_info, list);

		SUPERMOUNT_BUG_ON(!sdi->dentry);

		subdent = sdi->dentry;
		sdi->dentry = 0;
		list_del_init(&sdi->list);
		d_drop(sdi->host);

		dput(subdent);
	}

	SUPERMOUNT_TRACE_LEAVE(sb);

}

/*
 * close all open inodes on subfs
 */
static void
supermount_clean_inodes(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct list_head *ptr, *n;
	struct supermount_inode_info *sii;
	struct inode *subi;

	SUPERMOUNT_TRACE_ENTER(sb);

	list_for_each_safe(ptr, n, &sbi->s_inodes) {
		sii = list_entry(ptr, struct supermount_inode_info, list);

		subi = sii->inode;
		sii->inode = NULL;
		list_del_init(&sii->list);

		SUPERMOUNT_BUG_LOCKED_ON(sb, !subi);

		iput(subi);

		/*
		 * FIXME
		 * Is it really possible for write-/readcount to ever
		 * be > 1?
		 */
		SUPERMOUNT_BUG_LOCKED_ON(sb, sii->writecount < 0);
		SUPERMOUNT_BUG_LOCKED_ON(sb, sii->readcount < 0);
		while (sii->writecount) {
			sii->writecount--;
			sbi->s_writecount--;
		}
		while (sii->readcount) {
			sii->readcount--;
			__subfs_put_read_access(sb);
		}
	}

	SUPERMOUNT_TRACE_LEAVE(sb);

}

/*
 * Cut off busy dentries. They go away when their owners die
 */
static void
supermount_drop_dentries(struct super_block *sb)
{
        struct dentry *this_parent;
        struct list_head *next;

        if (!sb->s_root)
                goto out;
        this_parent = sb->s_root;

	spin_lock(&dcache_lock);
repeat:
        next = this_parent->d_subdirs.next;
resume:
        while(next != &this_parent->d_subdirs) {
                struct list_head *tmp = next;
                struct dentry *dentry = list_entry(tmp, struct dentry, d_child);

                next = tmp->next;

                list_del_init(&dentry->d_hash);

                if (!list_empty(&dentry->d_subdirs)) {
                        this_parent = dentry;
                        goto repeat;
                }
        }

        if (this_parent != sb->s_root) {
                next = this_parent->d_child.next;
                this_parent = this_parent->d_parent;
                goto resume;
        }
	spin_unlock(&dcache_lock);
out:
        return;
}


/*
 * FIXME
 * this should be split in subfs_cleanup and subfs_umount proper
 */
void
subfs_umount(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct vfsmount *mnt = subfs_mnt(sb);

	SUPERMOUNT_TRACE_ENTER(sb);

	sbi->s_undermount = NULL;
	sbi->s_generation++;

	/*
	 * we used to did shrink_dcache here. It compicates locking
	 * (clear_inode is called under sbi->sem thus requiring either
	 * recursive lock or separate lock just for inode list).
	 * This is not needed any more to ensure subfs can be umounted
	 * so we let dentries die and rely on dentry_revalidate to
	 * reject stale dentries
	 */

	if (sbi->s_opencount > 0)
		supermount_warning(sb, "busy files on changed media");

	supermount_clean_files(sb);
	supermount_clean_dentries(sb);
	supermount_clean_inodes(sb);

#if 0
	supermount_drop_dentries(sb);
#endif

	/*
	 * all files are expected to be closed
	 */
	SUPERMOUNT_BUG_ON(sbi->s_writecount);
	SUPERMOUNT_BUG_ON(sbi->s_opencount);

	/* paranoid check for double-mounting */
	SUPERMOUNT_BUG_ON(atomic_read(&mnt->mnt_sb->s_active) != 1);

	mntput(mnt);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

static inline int
subfs_remount_ro(struct super_block *sb)
{
	struct super_block *subsb = subfs_sb(sb);
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = do_remount_sb(subsb, subsb->s_flags | MS_RDONLY, NULL);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static inline int
subfs_remount_rw(struct super_block *sb)
{
	struct super_block *subsb = subfs_sb(sb);
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	rc = do_remount_sb(subsb, subsb->s_flags & ~MS_RDONLY, NULL);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static struct vfsmount *
subfs_real_mount(struct super_block *sb, char *type)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct vfsmount *mnt;

	if (sbi->s_data_copy)
		strcpy(sbi->s_data, sbi->s_data_copy);

	mnt = do_kern_mount(type,
			    sb->s_flags | MS_MGC_VAL | MS_SUPERMOUNTED,
			    sbi->s_devname, sbi->s_data);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return mnt;
}

static int
subfs_real_mount2(struct super_block *sb, char *type)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct vfsmount *mnt;
	struct supermount_dentry_info *root_sdi;
	struct supermount_inode_info *root_sii;
	int rc = 0;

	SUPERMOUNT_TRACE_ENTER(sb);

	sb->s_flags = sbi->s_mflags;

	mnt = subfs_real_mount(sb, type);

	if (IS_ERR(mnt) && PTR_ERR(mnt) == -EROFS) {
		sb->s_flags |= MS_RDONLY;
		mnt = subfs_real_mount(sb, type);
	}
	rc = PTR_ERR(mnt);
	if (IS_ERR(mnt))
		goto out;

	sbi->s_undermount = mnt;

	if (!(sb->s_flags & MS_RDONLY)) {
		rc = subfs_remount_ro(sb);
		if (rc)
			goto mntput;
	}

	root_sdi = supermount_d(sb->s_root);
	SUPERMOUNT_BUG_LOCKED_ON(sb, root_sdi->dentry);
	root_sdi->dentry = dget(mnt->mnt_root);
	root_sdi->generation = sbi->s_generation;
	list_add(&root_sdi->list, &sbi->s_dentries);
	if (list_empty(&sb->s_root->d_hash))
		d_rehash(sb->s_root);

	root_sii = supermount_i(sb->s_root->d_inode);
	SUPERMOUNT_BUG_LOCKED_ON(sb, root_sii->inode);
	root_sii->inode = igrab(mnt->mnt_root->d_inode);
	root_sii->generation = sbi->s_generation;
	list_add(&root_sii->list, &sbi->s_inodes);

	rc = 0;
	goto out;

	/*
	 * error clean up
	 */
mntput:
	mntput(mnt);
	sbi->s_undermount = 0;
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

/*
 * vfat and msdos do not have any valididty checks for superblock and must
 * be the last entries!
 */
static char *types_in_order[] = {
	"udf",
	"iso9660",
	"ext2",
	"vfat",
	"msdos",
	NULL
};

/*
 * Error values from mount
 *   ENOENT     - no device file (quite possible with devfs)
 *   ENXIO      - device does not exist
 *   ENOMEDIUM  - no medium inserted (surprise, surprise :)
 *   EBUSY      - attempt to mount on already mounted device
 *                we specifically disallow it even when both
 *                file system and mode (ro/rw) are the same
 */
static int
subfs_mount(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	int retval = -ENODEV;
	char **type = types_in_order;

	SUPERMOUNT_TRACE_ENTER(sb);

	if (sbi->s_type && strcmp(sbi->s_type, "auto"))
		retval = subfs_real_mount2(sb, sbi->s_type);
	else
		while (*type && retval
		       && retval != -ENXIO
		       && retval != -ENOMEDIUM
		       && retval != -ENOENT
		       && retval != -EBUSY) {
			retval = subfs_real_mount2(sb, *type);
			type++;
		}

	if (!retval)
		supermount_unlock_door(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return retval;
}

static int
__subfs_check_disk_change(struct super_block *sb)
{
	struct super_block *subsb;
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	subsb = subfs_sb(sb);
	rc = atomic_read(&subsb->s_media_changed);
	if (!rc)
		rc = check_disk_change(subsb->s_dev);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

/*
 * this must really be called subfs_active ...
 */
int
subfs_check_disk_change(struct super_block *sb)
{
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (subfs_is_active(sb))
		rc = __subfs_check_disk_change(sb);
	else
		rc = 1;
	subfs_unlock(sb);


	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
check_and_remount_subfs(struct super_block *sb)
{
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	if (subfs_is_mounted(sb)) {
		if (__subfs_check_disk_change(sb))
			subfs_umount(sb);
		else {
			rc = 0;
			goto out;
		}
	}

	rc = subfs_mount(sb);

out:
	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static void
__subfs_get_read_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	SUPERMOUNT_TRACE_ENTER(sb);

	SUPERMOUNT_BUG_LOCKED_ON(sb, sbi->s_opencount < 0);

	if (!sbi->s_opencount++)
		supermount_lock_door(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

int
subfs_get_read_file(struct super_block *sb, struct file *file)
{
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (is_file_obsolete(file))
		rc = -ESTALE;
	else {
		__subfs_get_read_access(sb);
		rc = 0;
	}
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;

}

int
subfs_get_read_inode(struct super_block *sb, struct inode *inode)
{
	int rc;
	
	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (is_inode_obsolete(inode))
		rc = -ESTALE;
	else {
		struct supermount_inode_info *sii = supermount_i(inode);

		SUPERMOUNT_BUG_LOCKED_ON(sb, sii->readcount < 0);

		sii->readcount++;
		__subfs_get_read_access(sb);
		rc = 0;
	}
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static void
__subfs_put_read_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	SUPERMOUNT_TRACE_ENTER(sb);

	SUPERMOUNT_BUG_LOCKED_ON(sb, sbi->s_opencount <= 0);

	if (!--sbi->s_opencount)
		supermount_unlock_door(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

void
subfs_put_read_file(struct super_block *sb, struct file *file)
{
	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (!is_file_obsolete(file))
		__subfs_put_read_access(sb);
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

void
subfs_put_read_inode(struct super_block *sb, struct inode *inode)
{
	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	/*
	 * and here it is not a bug unfortunately.
	 * */
	if (!is_inode_obsolete(inode)) {
		struct supermount_inode_info *sii = supermount_i(inode);

		SUPERMOUNT_BUG_LOCKED_ON(sb, sii->readcount <= 0);

		sii->readcount--;
		__subfs_put_read_access(sb);
	}
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

/*
 * Try to put the filesystem into writable mode for the duration of
 * an open(O_RDWR) or an inode operation requiring write access.  
 * Note that when a file is finally unlinked (i_nlink == 0) we put
 * write access in clear_inode because some filesystems are known
 * to refuse to truncate a file if they see ro file system
 */

static int
__subfs_get_write_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	SUPERMOUNT_BUG_LOCKED_ON(sb, sbi->s_writecount < 0);

	rc = 0;
	if (!sbi->s_writecount) {

		if (sb->s_flags & MS_RDONLY)
			rc = -EROFS;
		else
			rc = subfs_remount_rw(sb);
	}

	if (!rc)
		sbi->s_writecount++;

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

int
subfs_get_write_file(struct super_block *sb, struct file *file)
{
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (is_file_obsolete(file))
		rc = -ESTALE;
	else
		rc = __subfs_get_write_access(sb);
	subfs_unlock(sb);


	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

int
subfs_get_write_inode(struct super_block *sb, struct inode *inode)
{
	int rc;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (is_inode_obsolete(inode))
		rc = -ESTALE;
	else {
		struct supermount_inode_info *sii = supermount_i(inode);

		SUPERMOUNT_BUG_LOCKED_ON(sb, sii->writecount < 0);

		rc = __subfs_get_write_access(sb);
		if (!rc)
			sii->writecount++;
	}
	subfs_unlock(sb);


	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

/*
 * Once we're done, release write access.
 */
static void
__subfs_put_write_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	SUPERMOUNT_TRACE_ENTER(sb);

	SUPERMOUNT_BUG_LOCKED_ON(sb, sbi->s_writecount <= 0);

	if (!--sbi->s_writecount) {
		/*
		 * no need to fsync, it is done automatically on remount
		 */
		int rc = subfs_remount_ro(sb);
		if (rc)
			supermount_error(sb, "failed to remount ro, error = %d", rc);
	}

	SUPERMOUNT_TRACE_LEAVE(sb);

}

void
subfs_put_write_file(struct super_block *sb, struct file *file)
{
	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (!is_file_obsolete(file))
		__subfs_put_write_access(sb);
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

void
subfs_put_write_inode(struct super_block *sb, struct inode *inode)
{
	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (!is_inode_obsolete(inode)) {
		struct supermount_inode_info *sii = supermount_i(inode);

		SUPERMOUNT_BUG_LOCKED_ON(sb, sii->writecount <= 0);

		__subfs_put_write_access(sb);
		sii->writecount--;
	}
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

struct vfsmount *
subfs_go_online(struct super_block *sb)
{
	int rc;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);

	rc = check_and_remount_subfs(sb);
	mnt = ERR_PTR(rc);
	if (!rc)
		mnt = mntget(subfs_mnt(sb));
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return mnt;
}

/*
 * we do not need inode here, there is nothing really to check for
 */
void
subfs_go_offline(struct super_block *sb, struct vfsmount *mnt)
{
	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);

	SUPERMOUNT_BUG_LOCKED_ON(sb, !mnt);

	mntput(mnt);

	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

struct vfsmount *
subfs_prevent_umount(struct super_block *sb, struct inode *inode)
{
	struct vfsmount *mnt = 0;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (!is_inode_obsolete(inode))
		mnt = mntget(subfs_mnt(sb));
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return mnt;
}

void
subfs_allow_umount(struct super_block *sb, struct vfsmount *mnt)
{
	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	SUPERMOUNT_BUG_LOCKED_ON(sb, !mnt);
	mntput(mnt);
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

struct vfsmount *
subfs_get_mnt(struct super_block *sb)
{
	struct vfsmount *mnt = 0;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (subfs_is_active(sb))
		mnt = mntget(subfs_mnt(sb));
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return mnt;
}

struct super_block *
subfs_get_sb(struct super_block *sb)
{
	struct super_block *subsb = 0;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (subfs_is_active(sb))
		subsb = subfs_sb(sb);
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return subsb;
}
