/*
 *  linux/fs/supermount/super_operations.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/super.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2, 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@@abc.cap.ru)
 *  Rewritten for kernel 2.4.21 (C) 2003 Andrey Borzenkov
 *                                       (arvidjaar@@mail.ru)
 *
 */

#define S_DBG_TRACE_CURRENT S_DBG_TRACE_SUPER
#include "supermount.h"

char supermount_debug_enable;
char supermount_fail_statfs;

static struct super_operations supermount_sops;

static void
supermount_write_inode(struct inode *inode, int sync)
{
	struct super_block *sb = inode->i_sb;
	struct super_block *subsb;
	struct inode *subi;
	struct vfsmount *submnt;

	SUPERMOUNT_TRACE_ENTER(sb);

	if (subfs_check_disk_change(sb))
		goto out;

	submnt = subfs_prevent_umount(sb, inode);
	if (!submnt)
		goto out;

	subi = get_subfs_inode(inode);
	if (IS_ERR(subi))
		goto allow_umount;

	subsb = subfs_get_sb(sb);
	if (!sb)
		goto put_subi;

	if (subsb->s_op && subsb->s_op->write_inode) {
		if (subfs_get_write_inode(sb, inode))
			goto put_subi;

		/*
		 * FIXME
		 * Should we spinlock on inode_lock here?
		 * Normally flags are (re-)set under spinlock ... but
		 * inode.c does not export it
		 *
		 * FIXME
		 * Do we need to clear DIRTY flag?
		 * sync_one does it ...
		 */
		subi->i_state |= I_LOCK;
		subsb->s_op->write_inode(subi, sync);
		subi->i_state &= ~I_LOCK;

		subfs_put_write_inode(sb, inode);
	}
put_subi:
	iput(subi);
allow_umount:
	subfs_allow_umount(sb, submnt);
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return;
}

static void
supermount_clear_inode(struct inode *inode)
{
	struct supermount_inode_info *sii;
	struct super_block *sb = inode->i_sb;
	struct inode *subi;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	sii = supermount_i(inode);
	list_del(&sii->list);
	subi = sii->inode;
	sii->inode = 0;
	SUPERMOUNT_BUG_LOCKED_ON(sb, (subi == 0) ^ is_inode_obsolete(inode));
	subfs_unlock(sb);


	if (subi) {
		/*
		 * cf. dentry.c
		 * hopefully this does not break writecount management.
		 * root inode cannot be actually deleted
		 */
		SUPERMOUNT_BUG_ON(subi->i_sb->s_root->d_inode != subi && atomic_read(&subi->i_count) != 1);

		iput(subi);

		while(sii->writecount)
			subfs_put_write_inode(sb, inode);
		while(sii->readcount)
			subfs_put_read_inode(sb, inode);
	}

	kfree(sii);

	SUPERMOUNT_TRACE_LEAVE(sb);

}

static struct supermount_sb_info *
create_sbi(struct super_block *sb)
{
	struct supermount_sb_info *sbi = kmalloc(sizeof (*sbi), GFP_KERNEL);

	if (!sbi)
		return NULL;

	memset(sbi, 0, sizeof (*sbi));

	sbi->s_default_mode = 0777 | S_IFDIR;
	sbi->s_mflags = sb->s_flags & MS_RMT_MASK;
	sbi->s_undermount = 0;
	sbi->s_type = sbi->s_devname = sbi->s_data = sbi->s_data_copy = 0;
	INIT_LIST_HEAD(&sbi->s_inodes);
	INIT_LIST_HEAD(&sbi->s_files);
	INIT_LIST_HEAD(&sbi->s_dentries);
	init_MUTEX(&sbi->sem);

	return sbi;
}

static void
free_sbi(struct supermount_sb_info *sbi)
{
	if (sbi->s_type)
		kfree(sbi->s_type);
	if (sbi->s_devname)
		kfree(sbi->s_devname);
	if (sbi->s_data)
		kfree(sbi->s_data);
	if (sbi->s_data_copy)
		kfree(sbi->s_data_copy);
	kfree(sbi);
}

/* Release the superblock and any resources it has reserved 
 * it is called with the sb locked
 */

static void
supermount_put_super(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	if (subfs_is_mounted(sb))
		subfs_umount(sb);
	subfs_unlock(sb);

	sb->s_dev = NODEV;

	SUPERMOUNT_TRACE_LEAVE(sb);

	sb->u.generic_sbp = 0;
	free_sbi(sbi);

}

static int
copy_option(char **option, const char *val)
{
	char *tmp;
	if (!val || !*val)
		return -EINVAL;
	tmp = kmalloc(1 + strlen(val), GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;
	strcpy(tmp, val);
	*option = tmp;
	return 0;
}

static int
parse_options(char *options, struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	char *this_char;
	char *value;
	int rc;

	if (!options)
		return 0;

	while ((this_char = options)) {
		if (!strncmp(this_char, "--", 2))
			this_char += 2;
		if (!*this_char)
			break;
		if (*this_char == ',') {
			/* An empty option, or the option "--",
			   introduces options to be passed through to
			   the subfs */
			return copy_option(&sbi->s_data, ++this_char);
		}
		if ((options = strchr(this_char, ',')))
			*options++ = 0;

		if ((value = strchr(this_char, '=')))
			*value++ = 0;

		if (!strcmp(this_char, "fs")) {
			rc = copy_option(&sbi->s_type, value);
			if (rc)
				return rc;
		} else if (!strcmp(this_char, "dev")) {
			rc = copy_option(&sbi->s_devname, value);
			if (rc)
				return rc;
		} else if (!strcmp(this_char, "debug")) {
			if (value)
				sbi->s_debug = simple_strtoul(value, 0, 0);
			else
				sbi->s_debug = S_DBG_DEBUG;
		} else if (!strcmp(this_char, "fail_statfs_until_mount")) {
			supermount_fail_statfs = 1;
		} else {
			supermount_error(sb, "Unrecognized mount option \"%s\"",
					 this_char);
			return -EINVAL;
		}
	}
	return 0;
}

static struct inode *
supermount_root_inode(struct super_block *s)
{
	struct inode *inode = supermount_create_inode(s);

	if (!inode)
		return NULL;

	inode->i_version = 0;
	inode->i_op = &supermount_root_iops;
	inode->i_fop = &supermount_dir_operations;

	return inode;
}

/* read_super: the main mount() entry point into the VFS layer. */
struct super_block *
supermount_read_super(struct super_block *sb, void *data, int silent)
{
	struct inode *root_inode;
	struct dentry *root;
	struct supermount_sb_info *sbi = create_sbi(sb);

	if (!sbi)
		goto fail_no_memory;

	sb->u.generic_sbp = sbi;

	if (parse_options((char *) data, sb))
		goto fail_parsing;

	if (!sbi->s_devname) {
		supermount_error(sb, "no dev=<device> option");
		goto fail_parsing;
	}

	if (!sbi->s_type)
		supermount_warning(sb, "no fs=<filesystem> option, assuming fs=auto");

	if (sbi->s_data) {
		sbi->s_data_copy = kmalloc(1 + strlen(sbi->s_data), GFP_KERNEL);
		if (!sbi->s_data_copy) {
			supermount_error(sb, "not enough memory");
			goto fail_parsing;
		}
		strcpy(sbi->s_data_copy, sbi->s_data);
	} else
		sbi->s_data_copy = NULL;

	sb->s_blocksize = 1024;
	sb->s_blocksize_bits = 10;
	sb->s_magic = SUPERMOUNT_SUPER_MAGIC;
	sb->s_op = &supermount_sops;

	root_inode = supermount_root_inode(sb);
	if (!root_inode)
		goto fail_allocating_root_inode;

	root = d_alloc_root(root_inode);
	if (!root)
		goto fail_allocating_root_dentry;
	if (supermount_init_dentry_info(root))
		goto fail_init_root_info;

	sb->s_root = root;
	sbi->s_generation++;

	return sb;

fail_init_root_info:
	dput(root);
fail_allocating_root_dentry:
	iput(root_inode);
fail_parsing:
fail_allocating_root_inode:
	free_sbi(sbi);

fail_no_memory:
	return NULL;

}

/*
 * called with lock_super held
 */

static void
supermount_write_super(struct super_block *sb)
{
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER(sb);

	if (subfs_check_disk_change(sb))
		goto out;

	if (subfs_is_rw(sb)) {
		struct super_block *subsb;

		/* FIXME */
		mnt = subfs_get_mnt(sb);
		if (!mnt)
			goto out;

		subsb = subfs_get_sb(sb);
		if (subsb && subsb->s_op && subsb->s_op->write_super)
			subsb->s_op->write_super(subsb);

		mntput(mnt);
	}
out:

	SUPERMOUNT_TRACE_LEAVE(sb);

	return;
}

static int
supermount_statfs(struct super_block *sb, struct statfs *buf)
{
	int rc = 0;
	struct super_block *subsb;
	struct vfsmount *mnt;

	SUPERMOUNT_TRACE_ENTER(sb);

	(void)subfs_check_disk_change(sb);

	mnt = subfs_get_mnt(sb);
	if (!mnt)
		goto out;

	subsb = subfs_get_sb(sb);
	if (subsb && subsb->s_op && subsb->s_op->statfs)
		rc = subsb->s_op->statfs(subsb, buf);

	mntput(mnt);
out:
	buf->f_type = SUPERMOUNT_SUPER_MAGIC;

	SUPERMOUNT_TRACE_LEAVE(sb);

	return rc;
}

static int
supermount_remount_fs(struct super_block *sb, int *flags, char *data)
{
	return -ENOSYS;
}

static struct super_operations supermount_sops = {
	.write_inode	= supermount_write_inode,
	.put_inode	= force_delete,
	.clear_inode	= supermount_clear_inode,
	.put_super	= supermount_put_super,
	.write_super	= supermount_write_super,
	.statfs		= supermount_statfs,
	.remount_fs	= supermount_remount_fs,
};
