#ifndef _SUPERMOUNT_I_H
#define _SUPERMOUNT_I_H

#include <linux/module.h>

#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/locks.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/smp_lock.h>

/*
 * The supermount superblock magic number
 */

#define SUPERMOUNT_SUPER_MAGIC	0x9fa1
#define SUPERMOUNT_VERSION		"1.1.0"

#define S_DBG_DEBUG			0x001
#define S_DBG_TRACE_DENTRY		0x002
#define S_DBG_TRACE_FILE		0x004
#define S_DBG_TRACE_FILEMAP		0x008
#define S_DBG_TRACE_MEDIACTL		0x010
#define S_DBG_TRACE_NAMEI		0x020
#define S_DBG_TRACE_SUBFS		0x040
#define S_DBG_TRACE_SUPER		0x080
#define S_DBG_TRACE_TRANSLATION		0x100
#define S_DBG_TRACE_UTILS		0x200

extern struct file_system_type supermount_fs_type;

#ifdef SUPERMOUNT_DEBUG

#define SUPERMOUNT_BUG_ON(x) BUG_ON(x)
#define SUPERMOUNT_BUG_LOCKED_ON(sb, x) \
do { \
	if (x) { \
		subfs_unlock(sb); \
		BUG(); \
	} \
} while (0)

#define supermount_debug(sb, args...) \
do { \
	struct supermount_sb_info *sbi = supermount_sbi(sb); \
	char *dev = sbi->s_devname ? sbi->s_devname : "unknown"; \
\
	if (supermount_dbg(sb, S_DBG_DEBUG)) { \
		printk("%sSUPERMOUNT DEBUG [dev=%s] <%s:%d> ", \
			KERN_DEBUG, dev, __FUNCTION__, __LINE__); \
		printk(args); \
		printk("\n"); \
	} \
} while(0)

#define SUPERMOUNT_TRACE_ENTER_ARGS(sb, args...) \
do { \
	struct supermount_sb_info *sbi = supermount_sbi(sb); \
	char *dev = sbi->s_devname ? sbi->s_devname : "unknown"; \
\
	if (supermount_dbg(sb, S_DBG_TRACE_CURRENT)) { \
		printk("%sSUPERMOUNT TRACE [dev=%s] ENTER %s ", \
			KERN_DEBUG, dev, __FUNCTION__); \
		printk(args); \
		printk("\n"); \
	} \
} while (0)

#define SUPERMOUNT_TRACE_LEAVE_ARGS(sb, args...) \
do { \
	struct supermount_sb_info *sbi = supermount_sbi(sb); \
	char *dev = sbi->s_devname ? sbi->s_devname : "unknown"; \
\
	if (supermount_dbg(sb, S_DBG_TRACE_CURRENT)) { \
		printk("%sSUPERMOUNT TRACE [dev=%s] LEAVE %s ", \
			KERN_DEBUG, dev, __FUNCTION__); \
		printk(args); \
		printk("\n"); \
	} \
} while (0)

#else /* SUPERMOUNT_DEBUG */
#define supermount_debug(f, a...) /**/
#define SUPERMOUNT_BUG_ON(x) do { } while (0)
#define SUPERMOUNT_BUG_LOCKED_ON(x) do { } while (0)

#define SUPERMOUNT_TRACE_ENTER_ARGS(sb, args...) do { } while (0)
#define SUPERMOUNT_TRACE_LEAVE_ARGS(sb, args...) do { } while (0)
#endif				/* SUPERMOUNT_DEBUG */

#define SUPERMOUNT_TRACE_ENTER(sb) SUPERMOUNT_TRACE_ENTER_ARGS(sb, " ")
#define SUPERMOUNT_TRACE_LEAVE(sb) SUPERMOUNT_TRACE_LEAVE_ARGS(sb, " ") 

#define supermount_warning(sb, ...) \
do { \
	struct supermount_sb_info *sbi = supermount_sbi(sb); \
	char *dev = sbi->s_devname ? sbi->s_devname : "unknown"; \
\
	printk("%sSUPERMOUNT WARNING [dev=%s] ", KERN_WARNING, dev); \
	printk(__VA_ARGS__); \
	printk("\n"); \
} while(0)

#define supermount_error(sb, ...) \
do { \
	struct supermount_sb_info *sbi = supermount_sbi(sb); \
	char *dev = sbi->s_devname ? sbi->s_devname : "unknown"; \
\
	printk("%sSUPERMOUNT ERROR [dev=%s] ", KERN_ERR, dev); \
	printk(__VA_ARGS__); \
	printk("\n"); \
} while(0)

/*
 * supermount super-block data in memory
 */
struct supermount_sb_info {
	mode_t s_default_mode;	/* Default mode for supermount root */
	int s_generation;
	unsigned long s_debug;
	char *s_type;		/* Type of fs to be sub-mounted */
	char *s_devname;	/* Where to mount the subfs from */
	int s_mflags;		/* Flags to pass when mounting subfs */
	char *s_data;		/* Data to pass when mounting subfs */
	char *s_data_copy;	/* copy of previous data.  FS
				   destroy that data, and we
				   need to mount several
				   times.  We could allocate
				   the memory in
				   supermount_mount, but then
				   we have more possible
				   errors */
	struct vfsmount *s_undermount;	/* Mount point for subfs */
	int s_opencount;	/* Refcount of subfs usage; it is mostly
				   subfs dentries count except for inode
				   operations where several subfs dentries
				   may be in use at the same time */
	int s_writecount;	/* Refcount of write accesses
				   on the filesystem */
	struct list_head s_inodes;	/* list of active inodes */
	struct list_head s_files;	/* list of active files */
	struct list_head s_dentries;	/* list of active dentries */
	struct semaphore sem;
};

static inline struct supermount_sb_info *
supermount_sbi(struct super_block *sb)
{
	struct supermount_sb_info *sbi;
	SUPERMOUNT_BUG_ON(!sb);
	SUPERMOUNT_BUG_ON(sb->s_type != &supermount_fs_type);

	sbi = (struct supermount_sb_info *) (sb->u.generic_sbp);
	SUPERMOUNT_BUG_ON(!sbi);

	return sbi;
}

static int
supermount_sb_generation(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	return sbi->s_generation;
}

static inline struct vfsmount *
subfs_mnt(struct super_block *sb)
{
	return supermount_sbi(sb)->s_undermount;
}

static inline struct super_block *
subfs_sb(struct super_block *sb)
{
	struct vfsmount *mnt = subfs_mnt(sb);

	if (mnt) return mnt->mnt_sb;

	return 0;
}

/* this is expected to run under sb->sem */
static inline int
subfs_is_mounted(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	return sbi->s_undermount != 0;
}

/* this is expected to run under sb->sem */
static inline int
subfs_is_active(struct super_block *sb)
{
	struct super_block *subsb;

	if (!subfs_is_mounted(sb))
		return 0;
	
	subsb = subfs_sb(sb);
	return !atomic_read(&subsb->s_media_changed);
}

/* this is expected to run under sb->sem */
static inline int
subfs_is_rw(struct super_block *sb)
{
	struct super_block *subsb;

	if (!subfs_is_mounted(sb))
		return 0;
	
	subsb = subfs_sb(sb);
	return !(subsb->s_flags & MS_RDONLY);
}

static inline void
subfs_lock(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	down(&sbi->sem);
}

static inline void
subfs_unlock(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	up(&sbi->sem);
}

/*
 * query debug flags set
 */
static inline int supermount_dbg(struct super_block *sb, unsigned long flags)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	return sbi->s_debug & flags;
}

/*
 * supermount inode info
 */

struct supermount_inode_info {
	struct list_head list;
	int generation;
  	struct inode *inode;
	struct inode *host;
	int readcount;
	int writecount;
};

static inline int
is_inode_supermounted(struct inode *inode)
{
	return inode && inode->i_sb && inode->i_sb->s_type == &supermount_fs_type;
}

static inline struct supermount_inode_info *
supermount_i(struct inode *inode)
{
	SUPERMOUNT_BUG_ON(!is_inode_supermounted(inode));
	SUPERMOUNT_BUG_ON(!inode->u.generic_ip);

	return inode->u.generic_ip;
}

static inline int
is_inode_obsolete(struct inode *inode)
{
	struct supermount_inode_info *sii = supermount_i(inode);

	return sii->generation != supermount_sb_generation(inode->i_sb);
}

static inline void
supermount_list_add_inode(struct inode *inode)
{
	struct supermount_inode_info *sii = supermount_i(inode);

	list_add(&(sii->list), &(supermount_sbi(inode->i_sb)->s_inodes));
}

/*
 * supermount dentry info
 */

struct supermount_dentry_info {
	struct list_head list;
	int generation;
	struct dentry *dentry;
	struct dentry *host;
};

static inline int
is_dentry_supermounted(struct dentry *dentry)
{
	return (dentry && dentry->d_sb && dentry->d_sb->s_type == &supermount_fs_type);
}

static inline struct supermount_dentry_info *
supermount_d(struct dentry *dentry)
{
	SUPERMOUNT_BUG_ON(!is_dentry_supermounted(dentry));
	SUPERMOUNT_BUG_ON(!dentry->d_fsdata);

	return  (struct supermount_dentry_info *)dentry->d_fsdata;
}

static inline int
is_dentry_obsolete(struct dentry *dentry)
{
	struct supermount_dentry_info *sdi = supermount_d(dentry);

	return sdi->generation != supermount_sb_generation(dentry->d_sb);
}

/*
 * Supermount file info 
 */

struct supermount_file_info {
	struct list_head	list;
	struct file *		host;
	int generation;
	struct file		*file;
	pid_t			owner;
	struct vm_operations_struct *vm_ops;
};

static inline int
is_file_supermounted(struct file *file)
{
	return file && file->f_dentry && is_dentry_supermounted(file->f_dentry);
}

static inline struct supermount_file_info *
supermount_f(struct file *file)
{
	SUPERMOUNT_BUG_ON(!is_file_supermounted(file));
	SUPERMOUNT_BUG_ON(!file->f_supermount);

	return  file->f_supermount;
}

static inline int
is_file_obsolete(struct file *file)
{
	struct supermount_file_info *sfi = supermount_f(file);

	return sfi->generation != supermount_sb_generation(file->f_dentry->d_sb);
}

/*
 * FIXME
 * Intent is to remount subfs rw before updating atime from superinode
 * This is called in those cases when atime is updated by VFS and not
 * by subfs
 * We rely on update_atime to do all checking and just sync
 * values if it is mounted read-write
 */
#define SUPERMOUNT_UPDATE_ATIME(sb, inode, subi) UPDATE_ATIME(subi)

#if 0
#define SUPERMOUNT_UPDATE_ATIME(sb, inode, subi) \
do { \
	if (!(sb->s_flags & MS_RDONLY)) \
		if (subfs_get_write_access(sb)) { \
			supermount_update_atime(inode, subi); \
			subfs_put_write_access(sb); \
		} \
} while (0)
#endif


/* subfs.c */
extern void subfs_umount(struct super_block *sb);
extern struct vfsmount *subfs_go_online(struct super_block *);
extern void subfs_go_offline(struct super_block *, struct vfsmount *);
extern int subfs_get_write_file(struct super_block *, struct file *);
extern void subfs_put_write_file(struct super_block *, struct file *);
extern int subfs_get_write_inode(struct super_block *, struct inode *);
extern void subfs_put_write_inode(struct super_block *, struct inode *);
extern int subfs_get_read_file(struct super_block *, struct file *);
extern void subfs_put_read_file(struct super_block *, struct file *);
extern int subfs_get_read_inode(struct super_block *, struct inode *);
extern void subfs_put_read_inode(struct super_block *, struct inode *);
extern int subfs_check_disk_change(struct super_block *);
extern struct vfsmount *subfs_prevent_umount(struct super_block *, struct inode *);
extern void subfs_allow_umount(struct super_block *, struct vfsmount *);
extern struct vfsmount *subfs_get_mnt(struct super_block *sb);
extern struct super_block *subfs_get_sb(struct super_block *sb);
/* translation.c */
extern struct dentry *get_subfs_dentry(struct dentry *dentry);
extern struct inode *get_subfs_inode(struct inode *inode);
extern struct dentry *new_subfs_dentry(struct dentry *parent, struct dentry *super);
extern struct file *get_subfs_file(struct file*);
/* mediactl.c */
extern void supermount_handle_door(struct super_block *, int);
static inline void
supermount_lock_door(struct super_block *sb)
{
	supermount_handle_door(sb, MEDIA_LOCK);
}
static inline void
supermount_unlock_door(struct super_block *sb)
{
	supermount_handle_door(sb, MEDIA_UNLOCK);
}
/* filemap.c */
extern int supermount_file_mmap(struct file *, struct vm_area_struct *);

/* namei.c */
extern struct inode *supermount_create_inode(struct super_block *s);
extern struct inode_operations supermount_root_iops;
extern struct inode_operations supermount_dir_iops;
extern struct inode_operations supermount_file_iops;
extern struct inode_operations supermount_symlink_iops;
/* super_operations.c */
extern struct super_block *supermount_read_super(struct super_block *,
						 void *, int);
/* file.c */
extern struct file_operations supermount_dir_operations;
extern struct file_operations supermount_file_operations;
/* dentry.c */
extern struct dentry_operations supermount_dops;
/* filemap.c */
extern struct vm_operations_struct supermount_vm_ops;

/* utils.c */
extern int supermount_init_file_info(struct file *);
extern int supermount_attach_subfile(struct file *, struct file *);
extern int supermount_init_dentry_info(struct dentry *);
extern void supermount_attach_subdent(struct dentry *, struct dentry *);

#endif				/* _SUPERMOUNT_I_H */
