/*
 *  linux/fs/supermount/translation.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/dir.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/dir.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *  Rewriten for kernel 2.4. (C) 2001 MandrakeSoft Inc.
 *                                    Juan Quintela (quintela@mandrakesoft.com)
 *  Rewritten for kernel 2.4.21 (C) 2003 Andrey Borzenkov
 *  					 (arvidjaar@mail.ru)
 *
 */

#define S_DBG_TRACE_CURRENT S_DBG_TRACE_TRANSLATION
#include "supermount.h"

/*
 * all functions here return either ENOMEDIUM ot ESTALE. The order
 * is voluntary but not arbitrary. I prefer ENOMEDIUM for the case
 * when medium is not there (or has just been ejected)
 */
struct inode *
get_subfs_inode(struct inode *inode)
{
	struct super_block *sb = inode->i_sb;
	struct inode *err;
	struct supermount_inode_info *sii;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);

	sii = supermount_i(inode);
	SUPERMOUNT_BUG_LOCKED_ON(sb, (sii->inode == 0) ^ is_inode_obsolete(inode));

	err = ERR_PTR(-ENOMEDIUM);
	if (!subfs_is_active(sb))
		goto out;

	err = ERR_PTR(-ESTALE);
	if (is_inode_obsolete(inode))
		goto out;

	err = igrab(sii->inode);
	SUPERMOUNT_BUG_LOCKED_ON(sb, !err);

out:
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return err;
}

struct dentry *
get_subfs_dentry(struct dentry *dentry)
{
	struct super_block *sb = dentry->d_sb;
	struct dentry *err;
	struct supermount_dentry_info *sdi;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);

	sdi = supermount_d(dentry);
	SUPERMOUNT_BUG_LOCKED_ON(sb, (sdi->dentry == 0) ^ is_dentry_obsolete(dentry));

	err = ERR_PTR(-ENOMEDIUM);
	if (!subfs_is_active(sb))
		goto out;

	err = ERR_PTR(-ESTALE);
	if (is_dentry_obsolete(dentry))
		goto out;

	err = dget(sdi->dentry);
	SUPERMOUNT_BUG_LOCKED_ON(sb, !err);
	SUPERMOUNT_BUG_LOCKED_ON(sb, !err->d_inode);

out:
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return err;
}

/*
 * convert negative dentry into positive or allocate new one
 */
struct dentry *
new_subfs_dentry(struct dentry *parent, struct dentry *superdent)
{
	struct super_block *sb = superdent->d_sb;
	struct dentry *subdent;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);

	subdent = d_lookup(parent, &superdent->d_name);
	SUPERMOUNT_BUG_LOCKED_ON(sb, subdent && subdent->d_inode);

	/*
	 * if it was negative dentry, drop it and pretend
	 * we just allocated it
	 */
	if (subdent)
		d_drop(subdent);
	else
		subdent = d_alloc(parent, &superdent->d_name);

	if (!subdent)
		subdent = ERR_PTR(-ENOMEM);

	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return subdent;
}

struct file *
get_subfs_file(struct file *file)
{
	struct super_block *sb = file->f_dentry->d_sb;
	struct supermount_file_info *sfi;
	struct file *err;

	SUPERMOUNT_TRACE_ENTER(sb);

	subfs_lock(sb);
	
	sfi = supermount_f(file);
	SUPERMOUNT_BUG_LOCKED_ON(sb, (sfi->file == 0) ^ is_file_obsolete(file));

	err = ERR_PTR(-ENOMEDIUM);
	if (!subfs_is_active(sb))
		goto out;

	err = ERR_PTR(-ESTALE);
	if (is_file_obsolete(file))
		goto out;

	err = sfi->file;
	SUPERMOUNT_BUG_LOCKED_ON(sb, !err->f_dentry);
	SUPERMOUNT_BUG_LOCKED_ON(sb, !err->f_dentry->d_inode);
	get_file(err);

out:
	subfs_unlock(sb);

	SUPERMOUNT_TRACE_LEAVE(sb);

	return err;
}
